import os
import warnings
from typing import Union, List, Tuple

import spacy
from spacy.language import Language
import numpy as np
from sklearn.metrics import pairwise_distances

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language.common import SklearnTransformerMixin


class SpacyLanguage(SklearnTransformerMixin):
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a spaCy language
    backend. This object is meant for retreival, not plotting.

    Arguments:
        nlp: name of the model to load, be sure that it's downloaded beforehand

    **Usage**:

    ```python
    > lang = SpacyLanguage("en_core_web_md")
    > lang['python']
    > lang[['python', 'snake', 'dog']]
    > lang = SpacyLanguage("en_trf_robertabase_lg")
    > lang['programming in [python]']
    ```
    """

    def __init__(self, nlp: Union[str, Language]):
        self.nlp = nlp
        if isinstance(nlp, str):
            self.model = spacy.load(nlp)
        elif isinstance(nlp, Language):
            self.model = nlp
        else:
            raise ValueError(
                "Language must be started with `str` or spaCy-language object."
            )
        # Remove the lexeme prob table if it exists and is empty
        if (
            hasattr(self.model.vocab, "lookups_extra")
            and self.model.vocab.lookups_extra.has_table("lexeme_prob")
            and len(self.model.vocab.lookups_extra.get_table("lexeme_prob")) == 0
        ):
            self.model.vocab.lookups_extra.remove_table("lexeme_prob")

    @classmethod
    def from_fasttext(cls, language, output_dir, vectors_loc=None, force=False):
        """
        Will load downloaded fasttext vectors. It will try to load from disk, but if there is no local
        spaCy model then we will first convert from the vec.gz file into a spaCy model. This
        is saved on disk and then loaded as a spaCy model.

        Important:
            The fasttext vectors are not given by this library.
            You can download the models [here](https://fasttext.cc/docs/en/crawl-vectors.html#models).
            Note that these files are large and loading them can take a long time.

        Arguments:
            language: name of the language so that spaCy can grab correct tokenizer (example: "en" for english)
            output_dir: directory to save spaCy model
            vectors_loc: file containing the fasttext vectors
            force: with this flag raised we will always recreate the model from the vec.gz file

        **Usage**:

        ```python
        > lang = SpacyLanguage.from_fasttext("nl", "/path/spacy/model", "~/Downloads/cc.nl.300.vec.gz")
        > lang = SpacyLanguage.from_fasttext("en", "/path/spacy/model", "~/Downloads/cc.en.300.vec.gz")
        ```
        """
        if not os.path.exists(output_dir):
            spacy.cli.init_model(
                lang=language, output_dir=output_dir, vectors_loc=vectors_loc
            )
        else:
            if force:
                spacy.cli.init_model(
                    lang=language, output_dir=output_dir, vectors_loc=vectors_loc
                )
        return SpacyLanguage(spacy.load(output_dir))

    @staticmethod
    def _check_query_format(query: str) -> bool:
        opening = sum(1 for c in query if c == "[")
        closing = sum(1 for c in query if c == "]")
        if opening > 1:
            raise ValueError("Only one opening bracket (`[`) is allowed")
        if closing > 1:
            raise ValueError("Only one closing bracket (`]`) is allowed")
        if opening != closing:
            raise ValueError("The brackets should be paired")
        return opening > 0

    def __getitem__(
        self, query: Union[str, List[str]]
    ) -> Union[Embedding, EmbeddingSet]:
        """
        Retreive a single embedding or a set of embeddings. Depending on the spaCy model
        the strings can support multiple tokens of text but they can also use the Bert DSL.
        See the Language Options documentation: https://rasahq.github.io/whatlies/tutorial/languages/#bert-style.

        Arguments:
            query: single string or list of strings

        **Usage**
        ```python
        > lang = SpacyLanguage("en_core_web_md")
        > lang['python']
        > lang[['python', 'snake']]
        > lang[['nobody expects', 'the spanish inquisition']]
        > lang = SpacyLanguage("en_trf_robertabase_lg")
        > lang['programming in [python]']
        ```
        """
        if isinstance(query, str):
            return self._get_embedding(query)
        return EmbeddingSet(*[self._get_embedding(q) for q in query])

    def _get_embedding(self, query: str) -> Embedding:
        has_brackets = self._check_query_format(query)
        if has_brackets:
            start_idx, end_idx = self._get_context_pos(query)
            clean_query = query.replace("[", "").replace("]", "")
            vec = self.model(clean_query)[start_idx:end_idx].vector
            return Embedding(query, vec)
        return Embedding(query, self.model(query).vector)

    def _get_context_pos(self, query: str) -> Tuple[int, int]:
        tokens = list(t.text for t in self.model.tokenizer(query))
        start_idx = tokens.index("[")
        end_idx = tokens.index("]")
        return start_idx, end_idx - 1

    def _prepare_queries(self, prob_limit, lower):
        self._load_vocab()
        queries = [w for w in self.model.vocab]
        if prob_limit is not None:
            queries = [w for w in queries if w.prob >= prob_limit]
        if lower:
            queries = [w for w in queries if w.is_lower]
        if len(queries) == 0:
            raise ValueError(
                f"No tokens left for this setting. Consider raising prob_limit={prob_limit}"
            )
        return queries

    def _load_vocab(self):
        """This method must always be called before iterting over `model.vocab`."""
        # Load all the vocab only if they have not been loaded yet.
        if len(self.model.vocab) < len(self.model.vocab.vectors.keys()):
            for orth in self.model.vocab.vectors:
                self.model.vocab[orth]

    def _calculate_distances(self, emb, queries, metric):
        vec = emb.vector
        vector_matrix = np.array([w.vector for w in queries])
        return pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)

    def embset_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        prob_limit=-15,
        lower=True,
        metric="cosine",
    ):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most simmilar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            prob_limit: likelihood limit that sets the subset of words to search
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [w[0] for w in self.score_similar(emb, n, prob_limit, lower, metric)]
        return EmbeddingSet({w.name: w for w in embs})

    def embset_proximity(
        self,
        emb: Union[str, Embedding],
        max_proximity: float = 0.1,
        prob_limit=-15,
        lower=True,
        metric="cosine",
    ):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] or embeddings that are within a proximity.

        Arguments:
            emb: query to use
            max_proximity: the number of items you'd like to see returned
            prob_limit: likelihood limit that sets the subset of words to search
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(prob_limit, lower)
        distances = self._calculate_distances(emb, queries, metric)
        return EmbeddingSet(
            {w: self[w] for w, d in zip(queries, distances) if d <= max_proximity}
        )

    def score_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        prob_limit=-15,
        lower=True,
        metric="cosine",
    ):
        """
        Retreive a list of (Embedding, score) tuples that are the most simmilar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            prob_limit: likelihood limit that sets the subset of words to search, to ignore set to `None`
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(prob_limit, lower)
        distances = self._calculate_distances(emb, queries, metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])

        if len(queries) < n:
            warnings.warn(
                f"We could only find {len(queries)} feasible words. Consider changing `prob_limit` or `lower`",
                UserWarning,
            )

        return [(self[q.text], float(d)) for q, d in by_similarity[:n]]



from typing import Union
from copy import deepcopy

import numpy as np
from sklearn.metrics import pairwise_distances

from whatlies.common import handle_2d_plot


class Embedding:
    """
    This object represents a word embedding. It contains a vector and a name.

    Arguments:
        name: the name of this embedding, includes operations
        vector: the numerical representation of the embedding
        orig: original name of embedding, is left alone

    Usage:

    ```python
    from whatlies.embedding import Embedding

    foo = Embedding("foo", [0.1, 0.3])
    bar = Embedding("bar", [0.7, 0.2])

    foo | bar
    foo - bar + bar
    ```
    """

    def __init__(self, name, vector, orig=None):
        self.orig = name if not orig else orig
        self.name = name
        self.vector = np.array(vector)

    def add_property(self, name, func):
        result = Embedding(name=self.name, vector=self.vector, orig=self.orig,)
        setattr(result, name, func(result))
        return result

    def __add__(self, other) -> "Embedding":
        """
        Add two embeddings together.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo + bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} + {other.name})"
        copied.vector = self.vector + other.vector
        return copied

    def __sub__(self, other):
        """
        Subtract two embeddings.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo - bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} - {other.name})"
        copied.vector = self.vector - other.vector
        return copied

    def __gt__(self, other):
        """
        Measures the size of one embedding to another one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo > bar
        ```
        """
        return (self.vector.dot(other.vector)) / (other.vector.dot(other.vector))

    def __rshift__(self, other):
        """
        Maps an embedding unto another one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo >> bar
        ```
        """
        copied = deepcopy(self)
        new_vec = (
            (self.vector.dot(other.vector))
            / (other.vector.dot(other.vector))
            * other.vector
        )
        copied.name = f"({self.name} >> {other.name})"
        copied.vector = new_vec
        return copied

    def __or__(self, other):
        """
        Makes one embedding orthogonal to the other one.

        Usage:

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo | bar
        ```
        """
        copied = deepcopy(self)
        copied.name = f"({self.name} | {other.name})"
        copied.vector = self.vector - (self >> other).vector
        return copied

    def __repr__(self):
        return f"Emb[{self.name}]"

    def __str__(self):
        return self.name

    @property
    def norm(self):
        """Gives the norm of the vector of the embedding"""
        return np.linalg.norm(self.vector)

    def distance(self, other, metric: str = "cosine"):
        """
        Calculates the vector distance between two embeddings.

        Arguments:
            other: the other embedding you're comparing against
            metric: the distance metric to use, the list of valid options can be found [here](https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html)

        **Usage**

        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [1.0, 0.0])
        bar = Embedding("bar", [0.0, 0.5])

        foo.distance(bar)
        foo.distance(bar, metric="euclidean")
        foo.distance(bar, metric="cosine")
        ```
        """
        return pairwise_distances([self.vector], [other.vector], metric=metric)[0][0]

    def plot(
        self,
        kind: str = "scatter",
        x_axis: Union[str, "Embedding"] = None,
        y_axis: Union[str, "Embedding"] = None,
        color: str = None,
        show_ops: bool = False,
        annot: bool = False,
    ):
        """
        Handles the logic to perform a 2d plot in matplotlib.

        Arguments:
            kind: what kind of plot to make, can be `scatter`, `arrow` or `text`
            x_axis: the x-axis to be used, must be given when dim > 2
            y_axis: the y-axis to be used, must be given when dim > 2
            color: the color of the dots
            show_ops: setting to also show the applied operations, only works for `text`
            annot: should the points be annotated

        **Usage**
        ```python
        from whatlies.embedding import Embedding

        foo = Embedding("foo", [0.1, 0.3])
        bar = Embedding("bar", [0.7, 0.2])

        foo.plot(kind="arrow", annot=True)
        bar.plot(kind="arrow", annot=True)
        ```
        """
        if len(self.vector) == 2:
            handle_2d_plot(
                self,
                kind=kind,
                color=color,
                show_operations=show_ops,
                xlabel=x_axis,
                ylabel=y_axis,
                annot=annot,
            )
            return self
        x_val = self > x_axis
        y_val = self > y_axis
        intermediate = Embedding(name=self.name, vector=[x_val, y_val], orig=self.orig)
        handle_2d_plot(
            intermediate,
            kind=kind,
            color=color,
            xlabel=x_axis.name,
            ylabel=y_axis.name,
            show_operations=show_ops,
            annot=annot,
        )
        return self



import warnings

import numpy as np
from typing import Union, List
from sklearn.metrics import pairwise_distances
from gensim.models import KeyedVectors


from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language.common import SklearnTransformerMixin


class GensimLanguage(SklearnTransformerMixin):
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a keyed vector file.
    These files are generated by [gensim](https://radimrehurek.com/gensim/models/word2vec.html).
    This object is meant for retreival, not plotting.

    Important:
        The vectors are not given by this library they must be download/created upfront.
        A potential benefit of this is that you can train your own embeddings using
        gensim and visualise them using this library.

        Here's a snippet that you can use to train your own (very limited) word2vec embeddings.

        ```
        from gensim.test.utils import common_texts
        from gensim.models import Word2Vec
        model = Word2Vec(common_texts, size=10, window=5, min_count=1, workers=4)
        model.wv.save("wordvectors.kv")
        ```

        Note that if a word is not available in the keyed vectors file then we'll assume
        a zero vector. If you pass a sentence then we'll add together the embeddings vectors
        of the seperate words.

    Arguments:
        keyedfile: name of the model to load, be sure that it's downloaded or trained beforehand

    **Usage**:

    ```python
    > from whatlies.language import GensimLanguage
    > lang = GensimLanguage("wordvectors.kv")
    > lang['computer']
    > lang = GensimLanguage("wordvectors.kv")
    > lang[['computer', 'human', 'dog']]
    ```
    """

    def __init__(self, keyedfile):
        self.kv = KeyedVectors.load(keyedfile)

    def __getitem__(self, query: Union[str, List[str]]):
        """
        Retreive a single embedding or a set of embeddings.

        Arguments:
            query: single string or list of strings

        **Usage**
        ```python
        > from whatlies.language import GensimLanguage
        > lang = GensimLanguage("wordvectors.kv")
        > lang['computer']
        > lang = GensimLanguage("wordvectors.kv")
        > lang[['computer', 'human', 'dog']]
        ```
        """
        if isinstance(query, str):
            if " " in query:
                return Embedding(
                    query, np.sum([self[q].vector for q in query.split(" ")], axis=0)
                )
            try:
                vec = np.sum([self.kv[q] for q in query.split(" ")], axis=0)
            except KeyError:
                vec = np.zeros(self.kv.vector_size)
            return Embedding(query, vec)
        return EmbeddingSet(*[self[tok] for tok in query])

    def _prepare_queries(self, lower):
        queries = [w for w in self.kv.vocab.keys()]
        if lower:
            queries = [w for w in queries if w.lower() == w]
        return queries

    def _calculate_distances(self, emb, queries, metric):
        vec = emb.vector
        vector_matrix = np.array([self[w].vector for w in queries])
        # there are NaNs returned, good to investigate later why that might be
        vector_matrix = np.array(
            [np.zeros(v.shape) if np.any(np.isnan(v)) else v for v in vector_matrix]
        )
        return pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)

    def score_similar(
        self, emb: Union[str, Embedding], n: int = 10, metric="cosine", lower=False,
    ) -> List:
        """
        Retreive a list of (Embedding, score) tuples that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(lower=lower)
        distances = self._calculate_distances(emb=emb, queries=queries, metric=metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])

        if len(queries) < n:
            warnings.warn(
                f"We could only find {len(queries)} feasible words. Consider changing `top_n` or `lower`",
                UserWarning,
            )

        return [(self[q], float(d)) for q, d in by_similarity[:n]]

    def embset_similar(
        self, emb: Union[str, Embedding], n: int = 10, lower=False, metric="cosine",
    ) -> EmbeddingSet:
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [
            w[0] for w in self.score_similar(emb=emb, n=n, lower=lower, metric=metric)
        ]
        return EmbeddingSet({w.name: w for w in embs})



import warnings

import numpy as np
from typing import Union, List
from sklearn.metrics import pairwise_distances

import fasttext
import fasttext.util

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet

from whatlies.language.common import SklearnTransformerMixin, HiddenPrints


class FasttextLanguage(SklearnTransformerMixin):
    """
    This object is used to lazily fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a fasttext language
    backend. This object is meant for retreival, not plotting.

    Important:
        The vectors are not given by this library they must be download upfront.
        You can find the download links [here](https://fasttext.cc/docs/en/crawl-vectors.html).
        Note: you'll want the `bin` file, **not** the `text` file.
        To train your own fasttext model see the guide [here](https://fasttext.cc/docs/en/python-module.html#word-representation-model).

    Warning:
        You could theoretically use fasttext to train your own models with this code;

        ```
        > import fasttext
        > model = fasttext.train_unsupervised('data.txt',
                                              model='cbow',
                                              dim=10)
        > model = fasttext.train_unsupervised('data.txt',
                                              model='skipgram',
                                              dim=20,
                                              epoch=20,
                                              lr=0.1,
                                              min_count=1)
        > lang = FasttextLanguage(model)
        > lang['python']
        > model.save_model("result/data-skipgram-20.bin")
        > lang = FasttextLanguage("result/data-skipgram-20.bin")
        ```

        But you need to be aware that the fasttext library from facebook has gone stale.
        Last update on pypi was June 2019. Our preferred usecase for it is to use the pretrained vectors.
        Note that you can also import these via spaCy but this requires a packaging step.

    Arguments:
        model: name of the model to load, be sure that it's downloaded or trained beforehand

    **Usage**:

    ```python
    > from whatlies.language import FasttextLanguage
    > lang = FasttextLanguage("cc.en.300.bin")
    > lang['python']
    > lang = FasttextLanguage("cc.en.300.bin", size=10)
    > lang[['python', 'snake', 'dog']]
    ```
    """

    def __init__(self, model, size=None):
        self.size = size
        # we have to use this class to prevent the warning hidden as a print statement from the fasttext lib
        with HiddenPrints():
            if isinstance(model, str):
                self.model = fasttext.load_model(model)
            elif isinstance(model, fasttext.FastText._FastText):
                self.model = model
            else:
                raise ValueError(
                    "Language must be started with `str` or fasttext.FastText._FastText object."
                )
        if self.size:
            fasttext.util.reduce_model(self.model, self.size)

    @staticmethod
    def _input_str_legal(string):
        if sum(1 for c in string if c == "[") > 1:
            raise ValueError("only one opener `[` allowed ")
        if sum(1 for c in string if c == "]") > 1:
            raise ValueError("only one opener `]` allowed ")

    def __getitem__(self, query: Union[str, List[str]]):
        """
        Retreive a single embedding or a set of embeddings. Depending on the spaCy model
        the strings can support multiple tokens of text but they can also use the Bert DSL.
        See the Language Options documentation: https://rasahq.github.io/whatlies/tutorial/languages/#bert-style.

        Arguments:
            query: single string or list of strings

        **Usage**
        ```python
        > lang = FasttextLanguage("cc.en.300.bin")
        > lang['python']
        > lang[['python'], ['snake']]
        > lang[['nobody expects'], ['the spanish inquisition']]
        ```
        """
        if isinstance(query, str):
            self._input_str_legal(query)
            vec = self.model.get_word_vector(query)
            return Embedding(query, vec)
        return EmbeddingSet(*[self[tok] for tok in query])

    def _prepare_queries(self, top_n, lower):
        queries = [w for w in self.model.get_words()]
        if lower:
            queries = [w for w in queries if w.is_lower]
        if top_n is not None:
            queries = queries[:top_n]
        if len(queries) == 0:
            raise ValueError(
                f"Language model has no tokens for this setting. Consider raising top_n={top_n}"
            )
        return queries

    def _calculate_distances(self, emb, queries, metric):
        vec = emb.vector
        vector_matrix = np.array([self.model.get_word_vector(w) for w in queries])
        return pairwise_distances(vector_matrix, vec.reshape(1, -1), metric=metric)

    def embset_proximity(
        self,
        emb: Union[str, Embedding],
        max_proximity: float = 0.1,
        top_n=20_000,
        lower=True,
        metric="cosine",
    ):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] or embeddings that are within a proximity.

        Arguments:
            emb: query to use
            max_proximity: the number of items you'd like to see returned
            top_n: likelihood limit that sets the subset of words to search
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(top_n, lower)
        distances = self._calculate_distances(emb, queries, metric)
        return EmbeddingSet(
            {w: self[w] for w, d in zip(queries, distances) if d <= max_proximity}
        )

    def embset_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        top_n=20_000,
        lower=False,
        metric="cosine",
    ):
        """
        Retreive an [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            top_n: likelihood limit that sets the subset of words to search
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens, note that the official english model only has lower case tokens

        Important:
            This method is incredibly slow at the moment without a good `top_n` setting due to
            [this bug](https://github.com/facebookresearch/fastText/issues/1040).

        Returns:
            An [EmbeddingSet][whatlies.embeddingset.EmbeddingSet] containing the similar embeddings.
        """
        embs = [w[0] for w in self.score_similar(emb, n, top_n, lower, metric)]
        return EmbeddingSet({w.name: w for w in embs})

    def score_similar(
        self,
        emb: Union[str, Embedding],
        n: int = 10,
        top_n=20_000,
        lower=False,
        metric="cosine",
    ):
        """
        Retreive a list of (Embedding, score) tuples that are the most similar to the passed query.

        Arguments:
            emb: query to use
            n: the number of items you'd like to see returned
            top_n: likelihood limit that sets the subset of words to search, to ignore set to `None`
            metric: metric to use to calculate distance, must be scipy or sklearn compatible
            lower: only fetch lower case tokens, note that the official english model only has lower case tokens

        Important:
            This method is incredibly slow at the moment without a good `top_n` setting due
            to [this bug](https://github.com/facebookresearch/fastText/issues/1040).

        Returns:
            An list of ([Embedding][whatlies.embedding.Embedding], score) tuples.
        """
        if isinstance(emb, str):
            emb = self[emb]

        queries = self._prepare_queries(top_n, lower)
        distances = self._calculate_distances(emb, queries, metric)
        by_similarity = sorted(zip(queries, distances), key=lambda z: z[1])

        if len(queries) < n:
            warnings.warn(
                f"We could only find {len(queries)} feasible words. Consider changing `top_n` or `lower`",
                UserWarning,
            )

        return [(self[q], float(d)) for q, d in by_similarity[:n]]



from typing import Union, List

import tensorflow_text
import tensorflow as tf
import tensorflow_hub as tfhub

from whatlies.embedding import Embedding
from whatlies.embeddingset import EmbeddingSet
from whatlies.language.common import SklearnTransformerMixin, HiddenPrints


class ConveRTLanguage(SklearnTransformerMixin):
    """
    This object is used to fetch [Embedding][whatlies.embedding.Embedding]s or
    [EmbeddingSet][whatlies.embeddingset.EmbeddingSet]s from a
    [ConveRT](https://github.com/PolyAI-LDN/polyai-models) model.
    This object is meant for retreival, not plotting.

    Important:
        This object will automatically download a large file if it is not cached yet.

        Also note that this language model does not contain a vocabulary, so it cannot be used
        to retreive similar tokens. Instead, you can create an embedding set by using this model
        and then perform similarity queries from there.

    Arguments:
        model_id: identifier used for loading the corresponding TFHub module, which could be one of `'convert`, `'convert-multi-context'` or `'convert-ubuntu'`.
            Each one of these correspond to a different model as described in [ConveRT manual](https://github.com/PolyAI-LDN/polyai-models#models).
        signature: the TFHub signature of the model, which could be one of `'default'`, `'encode_context'`, `'encode_response'` or `'encode_sequence'`.
            Note that `'encode_context'` is not currently supported with `'convert-multi-context'` or `'convert-ubuntu'` models.

    **Usage**:

    ```python
    > from whatlies.language import ConveRTLanguage
    > lang = ConveRTLanguage()
    > lang['bank']
    > lang = ConveRTLanguage(model_id='convert-multi-context', signature='encode_sequence')
    > lang[['bank of the river', 'money on the bank', 'bank']]
    ```
    """

    MODEL_URL = {
        "convert": "http://models.poly-ai.com/convert/v1/model.tar.gz",
        "convert-multi-context": "http://models.poly-ai.com/multi_context_convert/v1/model.tar.gz",
        "convert-ubuntu": "http://models.poly-ai.com/ubuntu_convert/v1/model.tar.gz",
    }

    MODEL_SIGNATURES = [
        "default",
        "encode_context",
        "encode_response",
        "encode_sequence",
    ]

    def __init__(self, model_id: str = "convert", signature: str = "default") -> None:
        if model_id not in self.MODEL_URL:
            raise ValueError(
                f"The `model_id` value should be one of {list(self.MODEL_URL.keys())}"
            )
        if signature not in self.MODEL_SIGNATURES:
            raise ValueError(
                f"The `signature` value should be one of {self.MODEL_SIGNATURES}"
            )
        if signature == "encode_context" and model_id in [
            "convert-multi-context",
            "convert-ubuntu",
        ]:
            raise NotImplementedError(
                "Currently 'encode_context' signature is not support with multi-context and ubuntu models."
            )
        self.model_id = model_id
        self.signature = signature

        with HiddenPrints():
            self.module = tfhub.load(self.MODEL_URL[self.model_id])
            self.model = self.module.signatures[self.signature]

    def __getitem__(
        self, query: Union[str, List[str]]
    ) -> Union[Embedding, EmbeddingSet]:
        """
        Retreive a single embedding or a set of embeddings.

        Arguments:
            query: single string or list of strings

        **Usage**

        ```python
        > from whatlies.language import ConveRTLanguage
        > lang = ConveRTLanguage()
        > lang['bank']
        > lang = ConveRTLanguage()
        > lang[['bank of the river', 'money on the bank', 'bank']]
        ```
        """
        if isinstance(query, str):
            query_tensor = tf.convert_to_tensor([query])
            encoding = self.model(query_tensor)
            if self.signature == "encode_sequence":
                vec = encoding["sequence_encoding"].numpy().sum(axis=1)[0]
            else:
                vec = encoding["default"].numpy()[0]
            return Embedding(query, vec)
        return EmbeddingSet(*[self[tok] for tok in query])



from .spacy_lang import SpacyLanguage
from .sense2vec_lang import Sense2VecLanguage
from .fasttext_lang import FasttextLanguage
from .countvector_lang import CountVectorLanguage
from .bpemblang import BytePairLanguage
from .bpemblang import BytePairLanguage as BytePairLang
from .convert_lang import ConveRTLanguage
from .gensim_lang import GensimLanguage
from .tfhub_lang import TFHubLanguage
from .hftransformers_lang import HFTransformersLanguage


__all__ = [
    "SpacyLanguage",
    "Sense2VecLanguage",
    "FasttextLanguage",
    "CountVectorLanguage",
    "BytePairLang",
    "BytePairLanguage",
    "GensimLanguage",
    "ConveRTLanguage",
    "TFHubLanguage",
    "HFTransformersLanguage",
]



from whatlies.transformers.pca import Pca
from whatlies.transformers.umap import Umap
from whatlies.transformers.noise import Noise
from whatlies.transformers.addrandom import AddRandom
from whatlies.transformers.tsne import Tsne
from whatlies.transformers.opentsne import OpenTsne
from whatlies.transformers.ivis import Ivis

__all__ = ["Pca", "Umap", "Noise", "AddRandom", "Tsne", "OpenTsne", "Ivis"]
